import pygame
import random

# Colors
white = (255, 255, 255)
green = (0, 255, 213)
darkBlue = (0, 130, 196)
darkGray = (34, 39, 43)

# Display resolution
displayWidth = 1280
displayHeight = 900

# Initialize window
pygame.init()
window = pygame.display.set_mode((displayWidth, displayHeight))
window.fill(white)
pygame.display.set_caption("Igra vješala")

def displayText(message, color, fontSize, xOffset, yOffset, window):
    black = (0, 0, 0)
    font = pygame.font.Font("Roboto-Regular.ttf", fontSize)
    # Shadow
    text = font.render(message, True, black)
    window.blit(text, text.get_rect(center=(xOffset+1, yOffset+1)))
    # Text
    text = font.render(message, True, color)
    window.blit(text, text.get_rect(center=(xOffset, yOffset)))
    pygame.display.update()

def drawHangingPole(window, displayWidth):
    darkGray = (34, 39, 43)
    # Draw hanging pole
    pygame.draw.line(window, darkGray, (displayWidth/2, 80), (displayWidth/2, 40), 6)
    pygame.draw.line(window, darkGray, (displayWidth/2-2, 40), (displayWidth/2+120+3, 40), 6)
    pygame.draw.line(window, darkGray, (displayWidth/2+120, 40), (displayWidth/2+120, 360), 6)
    pygame.draw.line(window, darkGray, (displayWidth/2+40, 360), (displayWidth/2+160, 360), 6)

# Dictionary
words = {"otorinolaringologija"}
file = open("hrvatski_rjecnik.txt", "r", encoding="utf8")
for row in file:
    if row.split("\t")[3] == "imenica\n":
        words.add(row.split("\t")[1].lower())
words = list(filter(lambda x: True if x.isalpha() else False, words))

# Lives left
lives = 6

# Pick a word
word = random.choice(words)
length = len(word)
print(word, length)

# Is character opened
opened = [False for i in range(0, length+1)]

# Display empty word
text = ' '.join('_' * length)
displayText(text, darkGray, int(displayWidth/16), displayWidth/2, displayHeight-100, window)
drawHangingPole(window, displayWidth)


def drawStick(lives, window, displayWidth, displayHeight):
    red = (224, 67, 67)
    darkBlue = (0, 130, 196)
    darkGray = (34, 39, 43)
    if lives == 6: return
    # Draw head
    pygame.draw.arc(window, darkGray, (displayWidth / 2 - 29, 80, 61, 61), 0, 360, 6)
    if lives == 5: return
    # Draw body
    pygame.draw.line(window, darkGray, (displayWidth/2, 140), (displayWidth/2, 240), 6)
    if lives == 4: return
    # Draw left arm
    pygame.draw.line(window, darkGray, (displayWidth/2-45, 150), (displayWidth/2, 165), 6)
    if lives == 3: return
    # Draw right arm
    pygame.draw.line(window, darkGray, (displayWidth/2, 165), (displayWidth/2+45, 150), 6)
    if lives == 2: return
    # Draw left leg
    pygame.draw.line(window, darkGray, (displayWidth/2-25, 275), (displayWidth/2, 240), 8)
    if lives == 1: return
    # Draw right leg
    pygame.draw.line(window, darkGray, (displayWidth/2, 240), (displayWidth/2+25, 275), 8)
    displayText("Izgubili ste!", red, int(displayWidth/20), displayWidth/2, displayHeight/1.8, window)
    displayText("Riječ: " + word, darkBlue, int(displayWidth/20), displayWidth/2, displayHeight/1.5, window)
    displayText("Pritisnite space za novu igru", darkBlue, int(displayWidth/24), displayWidth/2, displayHeight/1.3, window)

# Game loop
run = True
while run:

    # Events
    for event in pygame.event.get():
        # Quit
        if event.type == pygame.QUIT:
            run = False
        # Pressed any key
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_SPACE:
                window.fill(white)
                # Lives left
                lives = 6
                # Reset game
                word = random.choice(words)
                length = len(word)
                print(word, length)
                # Is character opened
                opened = [False] * length
                # Display empty word
                text = ' '.join('_' * length)
                displayText(text, darkGray, int(displayWidth/16), displayWidth/2, displayHeight-100, window)
                drawHangingPole(window, displayWidth)
            # If key pressed is alphabet character
            if event.unicode.isalpha():
                # Check character
                correct = word.count(event.unicode.lower())
                if not correct:
                    lives -= 1
                for i in range(0, length):
                    if word[i] == event.unicode.lower():
                        opened[i] = True
                # Update word
                window.fill(white)
                drawHangingPole(window, displayWidth)
                drawStick(lives, window, displayWidth, displayHeight)
                text = ' '.join(list(map(lambda x, i: x if opened[i] else '_', word, range(0, length))))
                displayText(text, darkGray, int(displayWidth/16), displayWidth/2, displayHeight-100, window)
                # Check winner
                countOpened = sum([1 if x else 0 for x in opened])
                if countOpened == length:
                    displayText("Pobjedili ste!", green, int(displayWidth/20), displayWidth/2, displayHeight/1.7, window)
                    displayText("Pritisnite space za novu igru", darkBlue, int(displayWidth/24), displayWidth/2, displayHeight/1.4, window)

    pygame.display.update()
    pygame.time.delay(5)

pygame.quit()
