import pygame
import time

# Colors
lightGray = (118, 133, 148)
darkGray = (39, 44, 48)

# Display resolution
displaySize = 1000

# Initialize window
pygame.init()
window = pygame.display.set_mode((displaySize, displaySize))
window.fill(darkGray)
pygame.display.set_caption("Tic-Tac-Toe")
font = pygame.font.Font("Roboto-Regular.ttf", int(displaySize/7))

# Board
boardSize = 3
numForWin = 3
board = [[' ' for _ in range(0, boardSize)] for _ in range(0, boardSize)]

# Draw empty cells
borderWidth = displaySize / ((boardSize + 1) * 9)
cellWidth = displaySize / (boardSize * 1.125)
cells = [[0 for _ in range(0, boardSize)] for _ in range(0, boardSize)]
for i in range(0, boardSize):
    for j in range(0, boardSize):
        cells[i][j] = pygame.draw.rect(window, lightGray, (borderWidth + j * (cellWidth + borderWidth), borderWidth + i * (cellWidth + borderWidth), cellWidth, cellWidth))


def drawX(i, j, window, displaySize, boardSize, cells):
    blue = (0, 170, 255)
    darkBlue = (0, 130, 196)
    borderWidth = displaySize/((boardSize+1)*9)
    cellWidth = displaySize/(boardSize*1.125)
    lineWidth = cellWidth/8

    topLeftPoint = (cells[i][j].x + borderWidth + lineWidth/2, cells[i][j].y + borderWidth)
    bottomRightPoint = (cells[i][j].x + cellWidth-borderWidth - lineWidth/2, cells[i][j].y + cellWidth-borderWidth)
    topRightPoint = (cells[i][j].x + cellWidth-borderWidth - lineWidth/2, cells[i][j].y + borderWidth)
    bottomLeftPoint = (cells[i][j].x + borderWidth + lineWidth/2, cells[i][j].y + cellWidth-borderWidth)
    # Outline
    pygame.draw.line(window, darkBlue, (topLeftPoint[0] - 3, topLeftPoint[1] - 3), (bottomRightPoint[0] + 3, bottomRightPoint[1] + 3), int(lineWidth + 3*2))
    pygame.draw.line(window, darkBlue, (topRightPoint[0] + 3, topRightPoint[1] - 3), (bottomLeftPoint[0] - 3, bottomLeftPoint[1] + 3), int(lineWidth + 3*2))
    # X
    pygame.draw.line(window, blue, topLeftPoint, bottomRightPoint, int(lineWidth))
    pygame.draw.line(window, blue, topRightPoint, bottomLeftPoint, int(lineWidth))
    pygame.display.update()


def drawO(i, j, window, displaySize, boardSize, cells):
    green = (0, 255, 213)
    darkGreen = (0, 148, 124)
    borderWidth = displaySize/((boardSize+1)*9)
    cellWidth = displaySize/(boardSize*1.125)
    # Outline
    pygame.draw.arc(window, darkGreen, (cells[i][j].x + borderWidth - 3, cells[i][j].y + borderWidth - 3, cellWidth - 2 * borderWidth + 3*2, cellWidth - 2 * borderWidth + 3*2), 0, 360, int(cellWidth / 10 + 3*2))
    # O
    pygame.draw.arc(window, green, (cells[i][j].x + borderWidth, cells[i][j].y + borderWidth, cellWidth-2*borderWidth, cellWidth-2*borderWidth), 0, 360, int(cellWidth/10))
    pygame.display.update()


def displayText(message, displaySize, window, font):
    white = (255, 255, 255)
    darkGray = (39, 44, 48)
    # Shadow
    text = font.render(message, True, darkGray)
    window.blit(text, text.get_rect(center=(displaySize/2+3, displaySize/2+3)))
    # Text
    text = font.render(message, True, white)
    window.blit(text, text.get_rect(center=(displaySize/2, displaySize/2)))
    pygame.display.update()


def resetGame():
    global cells
    global board
    # Refill window
    window.fill(darkGray)
    # Redraw empty cells
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            cells[i][j] = pygame.draw.rect(window, lightGray, (borderWidth + j * (cellWidth + borderWidth), borderWidth + i * (cellWidth + borderWidth), cellWidth, cellWidth))
    # Reset board and board empty spaces
    board = [[' ' for _ in range(0, boardSize)] for _ in range(0, boardSize)]


def getLimit(direction, i, j, boardSize):
    if direction == 'row':
        return boardSize-j
    elif direction == 'column':
        return boardSize-i
    elif direction == 'main_diagonal':
        return boardSize-max(i, j)
    elif direction == 'secondary_diagonal':
        return min(i+1, boardSize-j)
    else:
        return None


def getInARow(direction):
    if direction == 'row':
        coef = (1, 0, 0, 1)  # [x*1 + k*0][y*0 + k*1] = [x][y+k]
    elif direction == 'column':
        coef = (0, 1, 1, 0)  # [x*0 + k*1][y*1 + k*0] = [x+k][y]
    elif direction == 'main_diagonal':
        coef = (1, 1, 1, 1)  # [x*1 + k*1][y*1 + k*1] = [x+k][y+k]
    elif direction == 'secondary_diagonal':
        coef = (1, -1, 1, 1)  # [x*1 + k*-1][y*1 + k*1] = [x-k][y+k]
    else:
        return None

    def countInARow(x, y, board, boardSize, numForWin):
        inARow = 0
        for i in range(0, boardSize):
            for j in range(0, boardSize):
                if board[i][j] == ' ' or getLimit(direction, i, j, boardSize) < numForWin:
                    continue
                posX = x
                posY = y
                if direction == 'row':
                    posY -= y
                elif direction == 'column':
                    posX -= x
                elif direction == 'main_diagonal':
                    posX -= min(x, y)
                    posY -= min(x, y)
                elif direction == 'secondary_diagonal':
                    while posX < boardSize-1 and posY > 0:
                        posX += 1
                        posY -= 1
                if getLimit(direction, posX, posY, boardSize) < numForWin:
                    continue
                inARow = 0
                player = board[x][y]
                for k in range(0, getLimit(direction, posX, posY, boardSize)):
                    if board[posX*coef[0] + k*coef[1]][posY*coef[2] + k*coef[3]] != player:
                        inARow = 0
                        continue
                    inARow += 1
                    if inARow == numForWin:
                        return inARow
        return inARow

    return countInARow


def checkWinner(x, y, board, boardSize, numForWin):
    # Row
    row = getInARow('row')
    if row(x, y, board, boardSize, numForWin) == numForWin:
        return board[x][y]

    # Column
    column = getInARow('column')
    if column(x, y, board, boardSize, numForWin) == numForWin:
        return board[x][y]

    # Main diagonal
    mainDiagonal = getInARow('main_diagonal')
    if mainDiagonal(x, y, board, boardSize, numForWin) == numForWin:
        return board[x][y]

    # Secondary diagonal
    secondaryDiagonal = getInARow('secondary_diagonal')
    if secondaryDiagonal(x, y, board, boardSize, numForWin) == numForWin:
        return board[x][y]

    # Draw
    emptySpaces = sum(list(map(lambda x: x.count(' '), board)))
    # emptySpaces = len(list(filter(lambda x: True if x == ' ' else False, [y for row in board for y in row])))
    if emptySpaces == 0:
        return 'draw'

    # Nobody won
    return 0


def heuristicEvaluate(board, isMaximizing, boardSize, numForWin):

    if isMaximizing:
        player = 'X'
    else:
        player = 'O'

    score = 0
    # Row
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ' or boardSize-j < numForWin:
                continue
            inARow = -1
            extraSpace = 0
            for k in range(0, boardSize-j):
                if board[i][j+k] != player and board[i][j+k] != ' ':
                    break
                if board[i][j+k] == ' ':
                    extraSpace += 1
                else:
                    inARow += 1
            if inARow >= 0 and inARow+1+extraSpace >= numForWin:
                score += 10**inARow
    # Column
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ' or boardSize-i < numForWin:
                continue
            inARow = -1
            extraSpace = 0
            for k in range(0, boardSize-i):
                if board[i+k][j] != player and board[i+k][j] != ' ':
                    break
                if board[i+k][j] == ' ':
                    extraSpace += 1
                else:
                    inARow += 1
            if inARow >= 0 and inARow+1+extraSpace >= numForWin:
                score += 10**inARow
    # Main diagonal
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ' or boardSize-max(i, j) < numForWin:
                continue
            inARow = -1
            extraSpace = 0
            for k in range(0, boardSize-max(i, j)):
                if board[i+k][j+k] != player and board[i+k][j+k] != ' ':
                    break
                if board[i+k][j+k] == ' ':
                    extraSpace += 1
                else:
                    inARow += 1
            if inARow >= 0 and inARow+1+extraSpace >= numForWin:
                score += 10**inARow
    # Secondary diagonal
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ' or min(i+1, boardSize-j) < numForWin:
                continue
            inARow = -1
            extraSpace = 0
            for k in range(0, min(i+1, boardSize-j)):
                if board[i-k][j+k] != player and board[i-k][j+k] != ' ':
                    break
                if board[i-k][j+k] == ' ':
                    extraSpace += 1
                else:
                    inARow += 1
            if inARow >= 0 and inARow+1+extraSpace >= numForWin:
                score += 10**inARow
    if isMaximizing:
        return score
    else:
        return -score


def minimax(x, y, alpha, beta, depth, isMaximizing, board, boardSize, numForWin):
    if boardSize > 3 and depth > 3:
        return heuristicEvaluate(board, isMaximizing, boardSize, numForWin)

    result = checkWinner(x, y, board, boardSize, numForWin)
    if result:
        if result == 'X':
            score = -100000000 + 5**depth
        elif result == 'O':
            score = 100000000 - 5**depth
        else:
            score = 0
        return score

    if isMaximizing:
        score = float("-inf")
        for i in range(0, boardSize):
            for j in range(0, boardSize):
                if score == 100000000 - 5**depth:
                    return score
                if board[i][j] == ' ':
                    board[i][j] = 'O'
                    score = max(score, minimax(i, j, alpha, beta, depth+1, False, board, boardSize, numForWin))
                    # print("---------------[ O ]---------------")
                    # out = str(board).split('], [')
                    # print(i, j, '],\n     ['.join(out), score)
                    board[i][j] = ' '
                    alpha = max(alpha, score)
                    if alpha >= beta:
                        return score
        return score
    else:
        score = float("inf")
        for i in range(0, boardSize):
            for j in range(0, boardSize):
                if score == -100000000 + 5**depth:
                    return score
                if board[i][j] == ' ':
                    board[i][j] = 'X'
                    score = min(score, minimax(i, j, alpha, beta, depth+1, True, board, boardSize, numForWin))
                    # print("---------------[ X ]---------------")
                    # out = str(board).split('], [')
                    # print(i, j, '],\n     ['.join(out), score)
                    board[i][j] = ' '
                    beta = min(beta, score)
                    if beta <= alpha:
                        return score
        return score

# Game loop
run = True
while run:

    # Events
    for event in pygame.event.get():
        # Quit
        if event.type == pygame.QUIT:
            run = False
        # Pressed space
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_SPACE:
                resetGame()
        # Mouse pressed
        if event.type == pygame.MOUSEBUTTONUP:
            # Get position
            position = pygame.mouse.get_pos()
            for i in range(0, boardSize):
                for j in range(0, boardSize):
                    # Has player clicked in any cell that is free?
                    if cells[i][j].collidepoint(position) and board[i][j] == ' ':
                        # Draw X and check if player won
                        drawX(i, j, window, displaySize, boardSize, cells)
                        board[i][j] = 'X'
                        winner = checkWinner(i, j, board, boardSize, numForWin)
                        if winner:
                            if winner == 'draw':
                                displayText("Draw!", displaySize, window, font)
                            else:
                                displayText("You " + ("win!" if winner == 'X' else "lose!"), displaySize, window, font)
                            # time.sleep(2)
                            # resetGame()
                        else:
                            # Computers turn, run minimax
                            bestScore = float("-inf")
                            for k in range(0, boardSize):
                                for l in range(0, boardSize):
                                    if board[k][l] == ' ':
                                        board[k][l] = 'O'
                                        before = time.time()
                                        score = minimax(k, l, float("-inf"), float("inf"), 0, False, board, boardSize, numForWin)
                                        after = time.time()
                                        print("\n---------------[ O ]---------------")
                                        out = str(board).split('], [')
                                        print(k, l, '],\n     ['.join(out), score)
                                        print('\nTime: ', after-before)
                                        board[k][l] = ' '
                                        if score > bestScore:
                                            bestScore = score
                                            i = k
                                            j = l
                            print("\nWaiting for player's turn...")
                            # Draw O and check if computer won
                            drawO(i, j, window, displaySize, boardSize, cells)
                            board[i][j] = 'O'
                            winner = checkWinner(i, j, board, boardSize, numForWin)
                            if winner:
                                if winner == 'draw':
                                    displayText("Draw!", displaySize, window, font)
                                else:
                                    displayText("You " + ("win!" if winner == 'X' else "lose!"), displaySize, window, font)

    pygame.display.update()
    pygame.time.delay(5)

pygame.quit()
