import pygame
import time

# Colors
blue = (0, 170, 255)
darkBlue = (0, 130, 196)
green = (0, 255, 213)
darkGreen = (0, 148, 124)
white = (255, 255, 255)
lightGray = (118, 133, 148)
gray = (70, 84, 94)
darkGray = (39, 44, 48)

# Display resolution
displaySize = 1000

# Initialize window
pygame.init()
window = pygame.display.set_mode((displaySize, displaySize))
window.fill(darkGray)
pygame.display.set_caption("Tic-Tac-Toe")
font = pygame.font.Font("Roboto-Regular.ttf", int(displaySize/7))

# Board
boardSize = 3
numForWin = 3
board = []
for _ in range(0, boardSize):
    temp = []
    for _ in range(0, boardSize):
        temp.append(' ')
    board.append(temp)

# Draw cells
borderWidth = displaySize/((boardSize+1)*9)
cellWidth = displaySize/(boardSize*1.125)
cells = [[0 for i in range(0, boardSize)] for i in range(0, boardSize)]
for i in range(0, boardSize):
    for j in range(0, boardSize):
        cells[i][j] = pygame.draw.rect(window, lightGray, (borderWidth + j * (cellWidth + borderWidth), borderWidth + i * (cellWidth + borderWidth), cellWidth, cellWidth))


def drawX(i, j):
    lineWidth = cellWidth/8
    topLeftPoint = (cells[i][j].x + borderWidth + lineWidth/2, cells[i][j].y + borderWidth)
    bottomRightPoint = (cells[i][j].x + cellWidth-borderWidth - lineWidth/2, cells[i][j].y + cellWidth-borderWidth)
    topRightPoint = (cells[i][j].x + cellWidth-borderWidth - lineWidth/2, cells[i][j].y + borderWidth)
    bottomLeftPoint = (cells[i][j].x + borderWidth + lineWidth/2, cells[i][j].y + cellWidth-borderWidth)
    # Outline
    pygame.draw.line(window, darkBlue, (topLeftPoint[0] - 3, topLeftPoint[1] - 3), (bottomRightPoint[0] + 3, bottomRightPoint[1] + 3), int(lineWidth + 3*2))
    pygame.draw.line(window, darkBlue, (topRightPoint[0] + 3, topRightPoint[1] - 3), (bottomLeftPoint[0] - 3, bottomLeftPoint[1] + 3), int(lineWidth + 3*2))
    # X
    pygame.draw.line(window, blue, topLeftPoint, bottomRightPoint, int(lineWidth))
    pygame.draw.line(window, blue, topRightPoint, bottomLeftPoint, int(lineWidth))
    pygame.display.update()


def drawO(i, j):
    # Outline
    pygame.draw.arc(window, darkGreen, (cells[i][j].x + borderWidth - 3, cells[i][j].y + borderWidth - 3, cellWidth - 2 * borderWidth + 3*2, cellWidth - 2 * borderWidth + 3*2), 0, 360, int(cellWidth / 10 + 3*2))
    # O
    pygame.draw.arc(window, green, (cells[i][j].x + borderWidth, cells[i][j].y + borderWidth, cellWidth-2*borderWidth, cellWidth-2*borderWidth), 0, 360, int(cellWidth/10))
    pygame.display.update()


def displayText(message):
    # Shadow
    text = font.render(message, True, darkGray)
    window.blit(text, text.get_rect(center=(displaySize/2 + 3, displaySize/2 + 3)))
    # Text
    text = font.render(message, True, white)
    window.blit(text, text.get_rect(center=(displaySize/2, displaySize/2)))
    pygame.display.update()


def resetGame():
    global cells
    global board
    # Refill window
    window.fill(darkGray)
    # Redraw cells
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            cells[i][j] = pygame.draw.rect(window, lightGray, (borderWidth + j * (cellWidth + borderWidth), borderWidth + i * (cellWidth + borderWidth), cellWidth, cellWidth))
    # Reset board
    board = []
    for _ in range(0, boardSize):
        temp = []
        for _ in range(0, boardSize):
            temp.append(' ')
        board.append(temp)


def checkWinner(x, y):
    # Row
    for j in range(0, boardSize):
        if board[x][j] == ' ' or boardSize-j < numForWin:
            continue
        inARow = 0
        player = board[x][j]
        for k in range(0, boardSize-j):
            if board[x][j+k] != player:
                break
            inARow += 1
            if inARow == numForWin:
                return player

    # Column
    for i in range(0, boardSize):
        if board[i][y] == ' ' or boardSize-i < numForWin:
            continue
        inARow = 0
        player = board[i][y]
        for k in range(0, boardSize-i):
            if board[i+k][y] != player:
                break
            inARow += 1
            if inARow == numForWin:
                return player

    # Main diagonal
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ' or boardSize-max(i, j) < numForWin:
                continue
            k = x - min(x, y)
            l = y - min(x, y)
            inARow = 0
            player = board[x][y]
            for m in range(0, boardSize-max(k, l)):
                if board[k+m][l+m] != player:
                    inARow = 0
                    continue
                inARow += 1
                if inARow == numForWin:
                    return player

    # Secondary diagonal
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ' or min(i+1, boardSize-j) < numForWin:
                continue
            k = x
            l = y
            while k < boardSize-1 and l > 0:
                k += 1
                l -= 1
            if min(k+1, boardSize-l) < numForWin:
                continue
            inARow = 0
            player = board[x][y]
            for m in range(0, min(k+1, boardSize-l)):
                if board[k-m][l+m] != player:
                    inARow = 0
                    continue
                inARow += 1
                if inARow == numForWin:
                    return player

    # Draw
    boardEmptySpaces = 0
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ':
                boardEmptySpaces += 1
    if boardEmptySpaces == 0:
        return 'draw'

    # Nobody won
    return 0


def heuristicEvaluate(board, isMaximizing):

    if isMaximizing:
        player = 'X'
    else:
        player = 'O'

    score = 0
    # Row
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ' or boardSize-j < numForWin:
                continue
            inARow = -1
            extraSpace = 0
            for k in range(0, boardSize-j):
                if board[i][j+k] != player and board[i][j+k] != ' ':
                    break
                if board[i][j+k] == ' ':
                    extraSpace += 1
                else:
                    inARow += 1
            if inARow >= 0 and inARow+1+extraSpace >= numForWin:
                score += 10**inARow
    # Column
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ' or boardSize-i < numForWin:
                continue
            inARow = -1
            extraSpace = 0
            for k in range(0, boardSize-i):
                if board[i+k][j] != player and board[i+k][j] != ' ':
                    break
                if board[i+k][j] == ' ':
                    extraSpace += 1
                else:
                    inARow += 1
            if inARow >= 0 and inARow+1+extraSpace >= numForWin:
                score += 10**inARow
    # Main diagonal
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ' or boardSize-max(i, j) < numForWin:
                continue
            inARow = -1
            extraSpace = 0
            for k in range(0, boardSize-max(i, j)):
                if board[i+k][j+k] != player and board[i+k][j+k] != ' ':
                    break
                if board[i+k][j+k] == ' ':
                    extraSpace += 1
                else:
                    inARow += 1
            if inARow >= 0 and inARow+1+extraSpace >= numForWin:
                score += 10**inARow
    # Secondary diagonal
    for i in range(0, boardSize):
        for j in range(0, boardSize):
            if board[i][j] == ' ' or min(i+1, boardSize-j) < numForWin:
                continue
            inARow = -1
            extraSpace = 0
            for k in range(0, min(i+1, boardSize-j)):
                if board[i-k][j+k] != player and board[i-k][j+k] != ' ':
                    break
                if board[i-k][j+k] == ' ':
                    extraSpace += 1
                else:
                    inARow += 1
            if inARow >= 0 and inARow+1+extraSpace >= numForWin:
                score += 10**inARow
    if isMaximizing:
        return score
    else:
        return -score


def minimax(x, y, alpha, beta, depth, isMaximizing, board):
    if boardSize > 3 and depth > 3:
        return heuristicEvaluate(board, isMaximizing)

    result = checkWinner(x, y)
    if result:
        if result == 'X':
            score = -100000000 + 5**depth
        elif result == 'O':
            score = 100000000 - 5**depth
        else:
            score = 0
        return score

    if isMaximizing:
        score = float("-inf")
        for i in range(0, boardSize):
            for j in range(0, boardSize):
                if score == 100000000 - 5**depth:
                    return score
                if board[i][j] == ' ':
                    board[i][j] = 'O'
                    score = max(score, minimax(i, j, alpha, beta, depth+1, False, board))
                    # print("---------------[ O ]---------------")
                    # out = str(board).split('], [')
                    # print(i, j, '],\n     ['.join(out), score)
                    board[i][j] = ' '
                    alpha = max(alpha, score)
                    if alpha >= beta:
                        return score
        return score
    else:
        score = float("inf")
        for i in range(0, boardSize):
            for j in range(0, boardSize):
                if score == -100000000 + 5**depth:
                    return score
                if board[i][j] == ' ':
                    board[i][j] = 'X'
                    score = min(score, minimax(i, j, alpha, beta, depth+1, True, board))
                    # print("---------------[ X ]---------------")
                    # out = str(board).split('], [')
                    # print(i, j, '],\n     ['.join(out), score)
                    board[i][j] = ' '
                    beta = min(beta, score)
                    if beta <= alpha:
                        return score
        return score


# Game
run = True
while run:

    # Events
    for event in pygame.event.get():
        # Quit
        if event.type == pygame.QUIT:
            run = False
        # Pressed space
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_SPACE:
                resetGame()
        # Mouse pressed
        if event.type == pygame.MOUSEBUTTONUP:
            # Get position
            position = pygame.mouse.get_pos()
            for i in range(0, boardSize):
                for j in range(0, boardSize):
                    # Has player clicked in any cell that is free?
                    if cells[i][j].collidepoint(position) and board[i][j] == ' ':
                        # Draw X and check if player won
                        drawX(i, j)
                        board[i][j] = 'X'
                        winner = checkWinner(i, j)
                        if winner:
                            if winner == 'draw':
                                displayText("Draw!")
                            else:
                                displayText("You " + ("win!" if winner == 'X' else "lose!"))
                            # time.sleep(2)
                            # resetGame()
                        else:
                            # Computers turn
                            bestScore = float("-inf")
                            for k in range(0, boardSize):
                                for l in range(0, boardSize):
                                    if board[k][l] == ' ':
                                        board[k][l] = 'O'
                                        before = time.time()
                                        score = minimax(k, l, float("-inf"), float("inf"), 0, False, board)
                                        after = time.time()
                                        print("\n---------------[ O ]---------------")
                                        out = str(board).split('], [')
                                        print(k, l, '],\n     ['.join(out), score)
                                        print('\nTime: ', after-before)
                                        board[k][l] = ' '
                                        if score > bestScore:
                                            bestScore = score
                                            i = k
                                            j = l
                            print("\nWaiting for player's turn...")
                            # Draw O and check if computer won
                            drawO(i, j)
                            board[i][j] = 'O'
                            winner = checkWinner(i, j)
                            if winner:
                                if winner == 'draw':
                                    displayText("Draw!")
                                else:
                                    displayText("You " + ("win!" if winner == 'X' else "lose!"))

    pygame.display.update()
    pygame.time.delay(5)

pygame.quit()
